package be.iminds.ilabt.jfed.experimenter_gui.events;

import be.iminds.ilabt.jfed.espec.bundle.ESpecBundle;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.time.Instant;

/**
 * User: twalcari
 * Date: 8/9/2016
 * Time: 16:54
 */
public class StartExperimentEvent {

    @Nonnull private final String sliceName;
    @Nullable private final String subAuthName;

    @Nullable private final Instant startTime;
    @Nonnull private final Instant expirationTime;

    @Nonnull private final RequestRspecSource requestRspecSource;
    @Nullable private final ESpecBundle eSpecBundle;

    private final boolean includeSshKeys;
    private final boolean requestSla;

    public StartExperimentEvent(@Nonnull String sliceName, @Nonnull String subAuthName,
                                @Nonnull Instant expirationTime,
                                @Nonnull RequestRspecSource requestRspecSource,
                                @Nullable ESpecBundle eSpecBundle,
                                boolean includeSshKeys, boolean requestSla) {
        this(sliceName, subAuthName, null, expirationTime, requestRspecSource, eSpecBundle, includeSshKeys, requestSla);
    }
    public StartExperimentEvent(@Nonnull String sliceName, @Nonnull String subAuthName,
                                @Nonnull Instant expirationTime,
                                @Nonnull RequestRspecSource requestRspecSource,
                                boolean includeSshKeys, boolean requestSla) {
        this(sliceName, subAuthName, null, expirationTime, requestRspecSource, includeSshKeys, requestSla);
    }

    public StartExperimentEvent(@Nonnull String sliceName, @Nullable String subAuthName,
                                @Nullable Instant startTime, @Nonnull Instant expirationTime,
                                @Nonnull RequestRspecSource requestRspecSource,
                                boolean includeSshKeys, boolean requestSla) {
        this(sliceName, subAuthName, startTime, expirationTime, requestRspecSource, null, includeSshKeys, requestSla);
    }

    public StartExperimentEvent(@Nonnull String sliceName, @Nullable String subAuthName,
                                @Nullable Instant startTime, @Nonnull Instant expirationTime,
                                @Nonnull RequestRspecSource requestRspecSource,
                                @Nullable ESpecBundle eSpecBundle,
                                boolean includeSshKeys, boolean requestSla) {
        this.sliceName = sliceName;
        this.subAuthName = subAuthName;
        this.startTime = startTime;
        this.expirationTime = expirationTime;
        this.requestRspecSource = requestRspecSource;
        this.eSpecBundle = eSpecBundle;
        this.includeSshKeys = includeSshKeys;
        this.requestSla = requestSla;
    }

    @Nonnull
    public String getSliceName() {
        return sliceName;
    }

    @Nullable
    public String getSubAuthName() {
        return subAuthName;
    }

    @Nonnull
    public Instant getExpirationTime() {
        return expirationTime;
    }

    @Nonnull
    public RequestRspecSource getRequestRspecSource() {
        return requestRspecSource;
    }

    @Nullable
    public ESpecBundle getESpecBundle() {
        return eSpecBundle;
    }

    @Nullable
    public Instant getStartTime() {
        return startTime;
    }

    public boolean isIncludeSshKeys() {
        return includeSshKeys;
    }

    public boolean isRequestSla() {
        return requestSla;
    }

    @Override
    public String toString() {
        return "StartExperimentEvent{" +
                "sliceName='" + sliceName + '\'' +
                ", subAuthName='" + subAuthName + '\'' +
                ", startTime=" + startTime +
                ", expirationTime=" + expirationTime +
                ", requestRspec=" + requestRspecSource.getRspecXmlString() +
                ", eSpecBundle=" + (eSpecBundle==null?null:eSpecBundle.toString()) +
                ", includeSshKeys=" + includeSshKeys +
                ", requestSla=" + requestSla +
                '}';
    }
}
